# coding: utf-8
"""General information about :class:`Tag<supervisely.annotation.tag.Tag>`"""

from __future__ import annotations
from typing import List, Optional, Dict
from copy import deepcopy
from supervisely.imaging.color import random_rgb, rgb2hex, hex2rgb, _validate_color
from supervisely.io.json import JsonSerializable
from supervisely.collection.key_indexed_collection import KeyObject
from supervisely._utils import take_with_default


class TagValueType:
    """
    Restricts Tag to have a certain value type.
    """

    NONE = "none"
    """"""
    ANY_NUMBER = "any_number"
    """"""
    ANY_STRING = "any_string"
    """"""
    ONEOF_STRING = "oneof_string"
    """"""


class TagMetaJsonFields:
    """
    Json fields for :class:`TagMeta<supervisely.annotation.tag_meta.TagMeta>`
    """

    ID = "id"
    """"""

    NAME = "name"
    """"""
    VALUE_TYPE = "value_type"
    """"""
    VALUES = "values"
    """"""
    COLOR = "color"
    """"""
    APPLICABLE_TYPE = "applicable_type"
    """"""
    HOTKEY = "hotkey"
    """"""
    APPLICABLE_CLASSES = "classes"
    """"""


class TagApplicableTo:
    """
    Defines Tag applicability only to images, objects or both.
    """

    ALL = "all"  # both images and objects
    """"""
    IMAGES_ONLY = "imagesOnly"
    """"""
    OBJECTS_ONLY = "objectsOnly"
    """"""


SUPPORTED_TAG_VALUE_TYPES = [
    TagValueType.NONE,
    TagValueType.ANY_NUMBER,
    TagValueType.ANY_STRING,
    TagValueType.ONEOF_STRING,
]
SUPPORTED_APPLICABLE_TO = [
    TagApplicableTo.ALL,
    TagApplicableTo.IMAGES_ONLY,
    TagApplicableTo.OBJECTS_ONLY,
]


class TagMeta(KeyObject, JsonSerializable):
    """
    General information about :class:`Tag<supervisely.annotation.tag>`. :class:`TagMeta<TagMeta>` object is immutable.

    :param name: Tag name.
    :type name: str
    :param value_type: Tag value type.
    :type value_type: str
    :param possible_values: List of possible values.
    :type possible_values: List[str], optional
    :param color: :class:`[R, G, B]` color, generates random color by default.
    :type color: List[int, int, int], optional
    :param sly_id: Tag ID in Supervisely server.
    :type sly_id: int, optional
    :param hotkey: Hotkey for Tag in annotation tool UI.
    :type hotkey: str, optional
    :param applicable_to: Defines applicability of Tag only to images, objects or both.
    :type applicable_to: str, optional
    :param applicable_classes: Defines applicability of Tag only to certain classes.
    :type applicable_classes: List[str], optional
    :raises: :class:`ValueError`, if color is not list, or doesn't have exactly 3 values
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # TagMeta
        meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)

        # TagMeta applicable only to Images example
        meta_cat = sly.TagMeta('cat', sly.TagValueType.NONE, applicable_to=sly.TagApplicableTo.IMAGES_ONLY)

        # TagMeta with string value applicable only to Objects example
        meta_breed = sly.TagMeta('breed', sly.TagValueType.ANY_STRING, applicable_to=sly.TagApplicableTo.OBJECTS_ONLY)

        # More complex TagMeta example
        # Create a list with possible values in order to use "ONEOF_STRING" value type
        coat_colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]
        # Note that "ONEOF_STRING" value type requires possible values, otherwise ValueError will be raised
        meta_coat_color = sly.TagMeta('coat color', sly.TagValueType.ONEOF_STRING, coat_colors, [255,120,0], hotkey="M", applicable_to=sly.TagApplicableTo.OBJECTS_ONLY, applicable_classes=["dog", "cat"])
    """

    def __init__(
        self,
        name: str,
        value_type: str,
        possible_values: Optional[List[str]] = None,
        color: Optional[List[int]] = None,
        sly_id: Optional[int] = None,
        hotkey: Optional[str] = None,
        applicable_to: Optional[str] = None,
        applicable_classes: Optional[List[str]] = None,
    ):
        if value_type not in SUPPORTED_TAG_VALUE_TYPES:
            raise ValueError(
                "value_type = {!r} is unknown, should be one of {}".format(
                    value_type, SUPPORTED_TAG_VALUE_TYPES
                )
            )

        self._name = name
        self._value_type = value_type
        self._possible_values = possible_values
        self._color = random_rgb() if color is None else deepcopy(color)
        self._sly_id = sly_id
        self._hotkey = take_with_default(hotkey, "")
        self._applicable_to = take_with_default(applicable_to, TagApplicableTo.ALL)
        self._applicable_classes = take_with_default(applicable_classes, [])
        if self._applicable_to not in SUPPORTED_APPLICABLE_TO:
            raise ValueError(
                "applicable_to = {!r} is unknown, should be one of {}".format(
                    self._applicable_to, SUPPORTED_APPLICABLE_TO
                )
            )

        if self._value_type == TagValueType.ONEOF_STRING:
            if self._possible_values is None:
                raise ValueError(
                    "TagValueType is ONEOF_STRING. List of possible values have to be defined."
                )
            if not all(isinstance(item, str) for item in self._possible_values):
                raise ValueError(
                    "TagValueType is ONEOF_STRING. All possible values have to be strings"
                )
        elif self._possible_values is not None:
            raise ValueError(
                "TagValueType is {!r}. possible_values variable have to be None".format(
                    self._value_type
                )
            )

        _validate_color(self._color)

    @property
    def name(self) -> str:
        """
        Name.

        :return: Name
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.ANY_STRING)
            print(meta_dog.name)
            # Output: 'dog'
        """
        return self._name

    def key(self) -> str:
        return self.name

    @property
    def value_type(self) -> str:
        """
        Value type. See possible value types in :class:`TagValueType<TagValueType>`.

        :return: Value type
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.ANY_STRING)
            meta_dog.value_type == sly.TagValueType.ANY_STRING # True

            print(meta_dog.value_type)
            # Output: 'any_string'
        """
        return self._value_type

    @property
    def possible_values(self) -> List[str]:
        """
        Possible values of object. This is a required field if object has "oneof_string" value type.

        :raise: :class:`ValueError` if list of possible values is not defined or TagMeta value_type is not "oneof_string".
        :return: List of possible values
        :rtype: :class:`List[str]`
        :Usage example:

         .. code-block:: python

            # List of possible values
            coat_colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]

            # TagMeta
            meta_coat_color = sly.TagMeta('coat color', sly.TagValueType.ONEOF_STRING, possible_values=coat_colors)

            print(meta_coat_color.possible_values)
            # Output: ['brown', 'white', 'black', 'red', 'chocolate', 'gold', 'grey']

            # Note that this is a required field if object has "oneof_string" value type.
            meta_coat_color = sly.TagMeta('coat color', sly.TagValueType.ONEOF_STRING)
            # Output: ValueError: TagValueType is ONEOF_STRING. List of possible values have to be defined.
        """
        return self._possible_values.copy() if self._possible_values is not None else None

    @property
    def color(self) -> List[int, int, int]:
        """
        :class:`[R,G,B]` color.

        :return: Color
        :rtype: :class:`List[int, int, int]`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE, color=[255,120,0])

            print(meta_dog.color)
            # Output: [255,120,0]
        """
        return self._color.copy()

    @property
    def sly_id(self) -> int:
        """
        Tag ID in Supervisely server.

        :return: ID
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE, sly_id=38584)

            print(meta_dog.sly_id)
            # Output: 38584
        """
        return self._sly_id

    @property
    def hotkey(self) -> str:
        """
        Hotkey for Tag in annotation tool UI.

        :return: Hotkey
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE, hotkey='M')

            print(meta_dog.hotkey)
            # Output: 'M'
        """
        return self._hotkey

    @property
    def applicable_to(self) -> str:
        """
        Tag applicability to objects, images, or both.

        :return: Applicability
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE, applicable_to=IMAGES_ONLY)

            print(meta_dog.applicable_to)
            # Output: 'imagesOnly'
        """
        return self._applicable_to

    @property
    def applicable_classes(self) -> List[str]:
        """
        Applicable classes.

        :returns: List of applicable classes
        :rtype: :class:`List[str]`
        :Usage example:

         .. code-block:: python

            # Imagine we have 2 ObjClasses in our Project
            class_car = sly.ObjClass(name='car', geometry_type='rectangle')
            class_bicycle = sly.ObjClass(name='bicycle', geometry_type='rectangle')

            # You can put a "string" with ObjClass name or use ObjClass.name
            meta_vehicle = sly.TagMeta('vehicle', sly.TagValueType.NONE, applicable_classes=["car", class_bicycle.name])

            print(meta_vehicle.applicable_classes)
            # Output: ['car', 'bicycle']
        """
        return self._applicable_classes

    def to_json(self) -> Dict:
        """
        Convert the TagMeta to a json dict. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            colors = ["brown", "white", "black", "red", "blue", "yellow", "grey"]
            meta_color = sly.TagMeta('Color',
                                    sly.TagValueType.ONEOF_STRING,
                                    possible_values=colors,
                                    color=[255, 120, 0],
                                    hotkey="M",
                                    applicable_classes=["car", "bicycle"])


            meta_color_json = meta_color.to_json()
            print(meta_color_json)
            # Output: {
            #     "name":"Color",
            #     "value_type":"oneof_string",
            #     "color":"#FF7800",
            #     "values":[
            #         "brown",
            #         "white",
            #         "black",
            #         "red",
            #         "blue",
            #         "yellow",
            #         "grey"
            #     ],
            #     "hotkey":"M",
            #     "applicable_type":"all",
            #     "classes":[
            #         "car",
            #         "bicycle"
            #     ]
            # }
        """
        jdict = {
            TagMetaJsonFields.NAME: self.name,
            TagMetaJsonFields.VALUE_TYPE: self.value_type,
            TagMetaJsonFields.COLOR: rgb2hex(self.color),
        }
        if self.value_type == TagValueType.ONEOF_STRING:
            jdict[TagMetaJsonFields.VALUES] = self.possible_values

        if self.sly_id is not None:
            jdict[TagMetaJsonFields.ID] = self.sly_id
        if self._hotkey is not None:
            jdict[TagMetaJsonFields.HOTKEY] = self.hotkey
        if self._applicable_to is not None:
            jdict[TagMetaJsonFields.APPLICABLE_TYPE] = self.applicable_to
        if self._applicable_classes is not None:
            jdict[TagMetaJsonFields.APPLICABLE_CLASSES] = self.applicable_classes

        return jdict

    @classmethod
    def from_json(cls, data: Dict) -> TagMeta:
        """
        Convert a json dict to TagMeta. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: TagMeta in json format as a dict.
        :type data: dict
        :return: TagMeta object
        :rtype: :class:`TagMeta<TagMeta>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            data = {
                "name":"Color",
                "value_type":"oneof_string",
                "color":"#FF7800",
                "values":[
                    "brown",
                    "white",
                    "black",
                    "red",
                    "blue",
                    "yellow",
                    "grey"
                ],
                "hotkey":"M",
                "applicable_type":"all",
                "classes":[
                    "car",
                    "bicycle"
                ]
            }

            meta_colors = sly.TagMeta.from_json(data)
        """
        if isinstance(data, str):
            return cls(name=data, value_type=TagValueType.NONE)
        elif isinstance(data, dict):
            name = data[TagMetaJsonFields.NAME]
            value_type = data[TagMetaJsonFields.VALUE_TYPE]
            values = data.get(TagMetaJsonFields.VALUES)
            color = data.get(TagMetaJsonFields.COLOR)
            if color is not None:
                color = hex2rgb(color)
            sly_id = data.get(TagMetaJsonFields.ID, None)

            hotkey = data.get(TagMetaJsonFields.HOTKEY, "")
            applicable_to = data.get(TagMetaJsonFields.APPLICABLE_TYPE, TagApplicableTo.ALL)
            applicable_classes = data.get(TagMetaJsonFields.APPLICABLE_CLASSES, [])

            return cls(
                name=name,
                value_type=value_type,
                possible_values=values,
                color=color,
                sly_id=sly_id,
                hotkey=hotkey,
                applicable_to=applicable_to,
                applicable_classes=applicable_classes,
            )
        else:
            raise ValueError("Tags must be dict or str types.")

    def add_possible_value(self, value: str) -> TagMeta:
        """
        Adds a new value to the list of possible values.

        :param value: New value that will be added to a list.
        :type value: str
        :raises: :class:`ValueError`, if object's value type is not "oneof_string" or already exists in a list
        :return: New instance of TagMeta
        :rtype: :class:`TagMeta<TagMeta>`
        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            #In order to add possible values, you must first initialize a variable where all possible values will be stored if it doesnt exist already
            colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]
            meta_coat_color = sly.TagMeta('coat color', sly.TagValueType.ONEOF_STRING, possible_values=colors, applicable_classes=["dog", "cat"])

            print(meta_coat_color.possible_values)
            # Output: ['brown', 'white', 'black', 'red', 'chocolate', 'gold', 'grey']

            #Now we can add new possible value to our TagMeta
            # Remember that TagMeta object is immutable, and we need to assign new instance of TagMeta to a new variable
            meta_coat_color = meta_coat_color.add_possible_value("bald (no coat)")

            print(meta_coat_color.possible_values)
            # Output: ['brown', 'white', 'black', 'red', 'chocolate', 'gold', 'grey', 'bald (no coat)']
        """
        if self.value_type is TagValueType.ONEOF_STRING:
            if value in self._possible_values:
                raise ValueError("Value {} already exists for tag {}".format(value, self.name))
            else:
                return self.clone(possible_values=[*self.possible_values, value])
        else:
            raise ValueError(
                "Tag {!r} has type {!r}. Possible value can be added only to oneof_string".format(
                    self.name, self.value_type
                )
            )

    def is_valid_value(self, value: str) -> bool:
        """
        Checks value against object value type to make sure that value is valid.

        :param value: Value to check.
        :type value: str
        :return: True if value is supported, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Initialize TagMeta
            meta_dog = sly.TagMeta('dog', sly.TagValueType.ANY_STRING)

            # Check what value type is in our Tagmeta
            print(meta_dog.value_type)
            # Output: 'any_string'

            # Our TagMeta has 'any_string' value type, it means only 'string' values will work with it
            # Let's check if value is valid for our TagMeta
            meta_dog.is_valid_value('Woof!')            # True
            meta_dog.is_valid_value(555)                # False

            # TagMetas with 'any_number' value type are compatible with 'int' and 'float' values
            meta_quantity = sly.TagMeta('quantity', sly.TagValueType.ANY_NUMBER)

            meta_quantity.is_valid_value('new string value') # False
            meta_quantity.is_valid_value(555)                # True
            meta_quantity.is_valid_value(3.14159265359)      # True
        """
        if self.value_type == TagValueType.NONE:
            return value is None
        elif self.value_type == TagValueType.ANY_NUMBER:
            return isinstance(value, (int, float))
        elif self.value_type == TagValueType.ANY_STRING:
            return isinstance(value, str)
        elif self.value_type == TagValueType.ONEOF_STRING:
            return isinstance(value, str) and (value in self._possible_values)
        else:
            raise ValueError("Unsupported TagValueType detected ({})!".format(self.value_type))

    def __eq__(self, other: TagMeta) -> bool:
        """
        Checks that 2 TagMetas are equal by their name, value type and possible values.

        :param other: TagMeta object.
        :type other: TagMeta
        :return: True if comparable objects are equal, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Let's create 2 identical TagMetas
            meta_lemon_1 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            meta_lemon_2 = sly.TagMeta('Lemon', sly.TagValueType.NONE)

            # and 1 different TagMeta and compare them to each other
            meta_cucumber = sly.TagMeta('Cucumber', sly.TagValueType.ANY_STRING)

            # Compare identical TagMetas
            meta_lemon_1 == meta_lemon_2      # True

            # Compare unidentical TagMetas
            meta_lemon_1 == meta_cucumber     # False
        """
        # TODO compare colors also here (need to check the usages and replace with is_compatible() where appropriate).
        return (
            isinstance(other, TagMeta)
            and self.name == other.name
            and self.value_type == other.value_type
            and self.possible_values == other.possible_values
        )

    def __ne__(self, other: TagMeta) -> bool:
        """
        Checks that 2 TagMetas are opposite.

        :param other: TagMeta object.
        :type other: TagMeta
        :return: True if comparable objects are not equal, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Let's create 2 identical TagMetas
            meta_lemon_1 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            meta_lemon_2 = sly.TagMeta('Lemon', sly.TagValueType.NONE)

            # and 1 different TagMeta and compare them to each other
            meta_cucumber = sly.TagMeta('Cucumber', sly.TagValueType.ANY_STRING)

            # Compare identical TagMetas
            meta_lemon_1 != meta_lemon_2      # False

            # Compare unidentical TagMetas
            meta_lemon_1 != meta_cucumber     # True
        """
        return not self == other

    def is_compatible(self, other: TagMeta) -> bool:
        """Check if the current TagMeta instance is compatible with another TagMeta instance."""
        if (
            not isinstance(other, TagMeta)
            or self.name != other.name
            or self.value_type != other.value_type
        ):
            return False

        if self.value_type == TagValueType.ONEOF_STRING:
            return sorted(self.possible_values) == sorted(other.possible_values)
        else:
            return True

    def clone(
        self,
        name: Optional[str] = None,
        value_type: Optional[str] = None,
        possible_values: Optional[List[str]] = None,
        color: Optional[List[int, int, int]] = None,
        sly_id: Optional[int] = None,
        hotkey: Optional[str] = None,
        applicable_to: Optional[str] = None,
        applicable_classes: Optional[List[str]] = None,
    ) -> TagMeta:
        """
        Clone makes a copy of TagMeta with new fields, if fields are given, otherwise it will use original TagMeta fields.

        :param name: Tag name.
        :type name: str
        :param value_type: Tag value type.
        :type value_type: str
        :param possible_values: List of possible values.
        :type possible_values: List[str], optional
        :param color: :class:`[R, G, B]` color, generates random color by default.
        :type color: List[int, int, int], optional
        :param sly_id: Tag ID in Supervisely server.
        :type sly_id: int, optional
        :param hotkey: Hotkey for Tag in annotation tool UI.
        :type hotkey: str, optional
        :param applicable_to: Defines applicability of Tag only to images, objects or both.
        :type applicable_to: str, optional
        :param applicable_classes: Defines applicability of Tag only to certain classes.
        :type applicable_classes: List[str], optional
        :return: New instance of TagMeta
        :rtype: :class:`TagMeta<TagMeta>`
        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            #Original TagMeta
            meta_dog_breed = sly.TagMeta('breed', sly.TagValueType.NONE)

            # TagMetas made of original TagMeta
            # Remember that TagMeta class object is immutable, and we need to assign new instance of TagMeta to a new variable
            A_breeds = ["Affenpinscher", "Afghan Hound", "Aidi", "Airedale Terrier", "Akbash Dog", "Akita"]
            meta_A_breed = meta_dog_breed.clone(value_type=sly.TagValueType.ONEOF_STRING, possible_values=A_breeds, hotkey='A')

            B_breeds = ["Basset Fauve de Bretagne", "Basset Hound", "Bavarian Mountain Hound", "Beagle", "Beagle-Harrier", "Bearded Collie"]
            meta_B_breed = meta_A_breed.clone(possible_values=B_breeds, hotkey='B')

            C_breeds = ["Cairn Terrier", "Canaan Dog", "Canadian Eskimo Dog", "Cane Corso", "Cardigan Welsh Corgi", "Carolina Dog"]
            meta_C_breed = meta_B_breed.clone(possible_values=C_breeds, hotkey='C')
        """
        return TagMeta(
            name=take_with_default(name, self.name),
            value_type=take_with_default(value_type, self.value_type),
            possible_values=take_with_default(possible_values, self.possible_values),
            color=take_with_default(color, self.color),
            sly_id=take_with_default(sly_id, self.sly_id),
            hotkey=take_with_default(hotkey, self.hotkey),
            applicable_to=take_with_default(applicable_to, self.applicable_to),
            applicable_classes=take_with_default(applicable_classes, self.applicable_classes),
        )

    def __str__(self):
        return (
            "{:<7s}{:<24} {:<7s}{:<13} {:<13s}{:<10} {:<13s}{:<10} {:<13s}{:<10} {:<13s}{}".format(
                "Name:",
                self.name,
                "Value type:",
                self.value_type,
                "Possible values:",
                str(self.possible_values),
                "Hotkey",
                self.hotkey,
                "Applicable to",
                self.applicable_to,
                "Applicable classes",
                self.applicable_classes,
            )
        )

    @classmethod
    def get_header_ptable(cls):
        """get_header_ptable"""
        return [
            "Name",
            "Value type",
            "Possible values",
            "Hotkey",
            "Applicable to",
            "Applicable classes",
        ]

    def get_row_ptable(self):
        """get_row_ptable"""
        return [
            self.name,
            self.value_type,
            self.possible_values,
            self.hotkey,
            self.applicable_to,
            self.applicable_classes,
        ]

    def _set_id(self, id: int):
        self._sly_id = id



# coding: utf-8

# docs
from __future__ import annotations

import base64
import io
import zlib
from typing import Optional

import cv2
import numpy as np
from PIL import Image

from supervisely.geometry.bitmap import Bitmap, _find_mask_tight_bbox
from supervisely.geometry.bitmap_base import BitmapBase
from supervisely.geometry.image_rotator import ImageRotator
from supervisely.geometry.point_location import PointLocation
from supervisely.imaging.image import read


class AlphaMask(Bitmap):
    """
    AlphaMask geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`AlphaMask<AlphaMask>` object is immutable.

    :param data: AlphaMask mask data. Must be a numpy array with values in range [0, 255].
    :type data: np.ndarray
    :param origin: :class:`PointLocation<supervisely.geometry.point_location.PointLocation>`: top, left corner of AlphaMask. Position of the AlphaMask within image.
    :type origin: PointLocation, optional
    :param sly_id: AlphaMask ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which AlphaMask belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created AlphaMask.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when AlphaMask was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when AlphaMask was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :param extra_validation: If True, additional validation is performed. Throws a ValueError if values of the data are not in the range [0, 255]. If True it will affect performance.
    :type extra_validation: bool, optional
    :raises: :class:`ValueError`, if data values are not in the range [0, 255].
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Create simple alpha mask:
        mask = np.array([[0, 0, 0, 0, 0],
                         [0, 50, 50, 50, 0],
                         [0, 50, 0, 50, 0],
                         [0, 50, 50, 50, 0],
                         [0, 0, 0, 0, 0]], dtype=np.uint8)

        figure = sly.AlphaMask(mask)

        origin = figure.origin.to_json()
        print(json.dumps(origin, indent=4))
        # Output: {
        #     "points": {
        #         "exterior": [
        #             [
        #                 1,
        #                 1
        #             ]
        #         ],
        #         "interior": []
        #     }

        # Create alpha mask from PNG image:
        img = sly.imaging.image.read(os.path.join(os.getcwd(), 'black_white.png'))
        mask = img[:, :, 3]
        figure = sly.AlphaMask(mask)
    """

    @staticmethod
    def geometry_name():
        """geometry_name"""
        return "alpha_mask"

    def __init__(
        self,
        data: np.ndarray,
        origin: Optional[PointLocation] = None,
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        extra_validation: Optional[bool] = True,
    ):
        if data.dtype != np.uint8:
            if data.dtype == np.bool:
                data = data.astype(np.uint8) * 255
            else:
                data = np.array(data, dtype=np.uint8)
            if extra_validation:
                if not np.all(np.isin(data, range(256))):
                    max_val = np.max(data)
                    min_val = np.min(data)
                    if max_val > 255 or min_val < 0:
                        raise ValueError(
                            f"Alpha mask data values must be in range [0, 255]. Instead got min: {min_val}, max: {max_val}."
                        )

        # Call base constructor first to do the basic dimensionality checks.
        BitmapBase.__init__(
            self,
            data=data,
            origin=origin,
            expected_data_dims=2,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

        if not np.any(data):
            raise ValueError(
                "Creating a alpha mask with an empty mask (no pixels set to True) is not supported."
            )
        data_tight_bbox = _find_mask_tight_bbox(self._data)
        self._origin = self._origin.translate(drow=data_tight_bbox.top, dcol=data_tight_bbox.left)
        self._data = data_tight_bbox.get_cropped_numpy_slice(self._data)

    def rotate(self, rotator: ImageRotator) -> AlphaMask:
        """
        Rotates current AlphaMask.

        :param rotator: :class:`ImageRotator<supervisely.geometry.image_rotator.ImageRotator>` for AlphaMask rotation.
        :type rotator: ImageRotator
        :return: AlphaMask object
        :rtype: :class:`AlphaMask<AlphaMask>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.geometry.image_rotator import ImageRotator

            height, width = ann.img_size
            rotator = ImageRotator((height, width), 25)
            # Remember that AlphaMask class object is immutable, and we need to assign new instance of AlphaMask to a new variable
            rotate_figure = figure.rotate(rotator)
        """
        full_img_mask = np.zeros(rotator.src_imsize, dtype=np.uint8)
        self.draw(full_img_mask, 255)
        # TODO this may break for one-pixel masks (it can disappear during rotation). Instead, rotate every pixel
        #  individually and set it in the resulting alpha mask.
        new_mask = rotator.rotate_img(full_img_mask, use_inter_nearest=True)
        return AlphaMask(data=new_mask)

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        """_draw_impl"""
        channels = bitmap.shape[2] if len(bitmap.shape) == 3 else 1
        non_zero_values = self.data > 0
        alpha = self.data / 255.0
        if channels >= 3:
            if isinstance(color, int):
                color = [color] * 3
            temp_mask = np.zeros(self.data.shape + (3,), dtype=np.uint8)
            temp_mask[non_zero_values] = color
            for i in range(3):
                canvas = self.to_bbox().get_cropped_numpy_slice(bitmap)[:, :, i]
                canvas[non_zero_values] = (canvas * (1 - alpha) + alpha * temp_mask[:, :, i])[
                    non_zero_values
                ]
        elif channels == 1:
            temp_mask = np.zeros(self.data.shape, dtype=np.uint8)
            temp_mask[non_zero_values] = color
            canvas = self.to_bbox().get_cropped_numpy_slice(bitmap)
            canvas[non_zero_values] = (canvas * (1 - alpha) + alpha * temp_mask)[non_zero_values]

    @property
    def area(self) -> float:
        """
        AlphaMask area.

        :return: Area of current AlphaMask
        :rtype: :class:`float`
        :Usage example:

         .. code-block:: python

            print(figure.area)
            # Output: 88101.0
        """
        return float(np.count_nonzero(self._data))

    @staticmethod
    def data_2_base64(mask: np.ndarray) -> str:
        """
        Convert numpy array to base64 encoded string.

        :param mask: Bool numpy array.
        :type mask: np.ndarray
        :return: Base64 encoded string
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            address = 'https://app.supervise.ly/'
            token = 'Your Supervisely API Token'
            api = sly.Api(address, token)

            # Get annotation from API
            meta_json = api.project.get_meta(PROJECT_ID)
            meta = sly.ProjectMeta.from_json(meta_json)
            ann_info = api.annotation.download(IMAGE_ID)
            ann = sly.Annotation.from_json(ann_info.annotation, meta)

            # Get AlphaMask from annotation
            for label in ann.labels:
                if type(label.geometry) == sly.AlphaMask:
                    figure = label.geometry

            encoded_string = sly.AlphaMask.data_2_base64(figure.data)
            print(encoded_string)
            # 'eJzrDPBz5+WS4mJgYOD19HAJAtLMIMwIInOeqf8BUmwBPiGuQPr///9Lb86/C2QxlgT5BTM4PLuRBuTwebo4hlTMSa44sKHhISMDuxpTYrr03F6gDIOnq5/LOqeEJgDM5ht6'
        """
        # img_pil = Image.fromarray(mask)
        img_pil = Image.fromarray(np.array(mask, dtype=np.uint8), mode="L")
        bytes_io = io.BytesIO()
        img_pil.save(bytes_io, format="PNG", transparency=0, optimize=0)
        bytes_enc = bytes_io.getvalue()
        return base64.b64encode(zlib.compress(bytes_enc)).decode("utf-8")

    @staticmethod
    def base64_2_data(s: str) -> np.ndarray:
        """
        Convert base64 encoded string to numpy array. Supports both compressed and uncompressed masks.

        :param s: Input base64 encoded string.
        :type s: str
        :return: numpy array
        :rtype: :class:`np.ndarray`
        :Usage example:

         .. code-block:: python

              import supervisely as sly

              encoded_string = 'eJzrDPBz5+WS4mJgYOD19HAJAtLMIMwIInOeqf8BUmwBPiGuQPr///9Lb86/C2QxlgT5BTM4PLuRBuTwebo4hlTMSa44sKHhISMDuxpTYrr03F6gDIOnq5/LOqeEJgDM5ht6'
              figure_data = sly.AlphaMask.base64_2_data(encoded_string)
              print(figure_data)

              uncompressed_string = 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsMAAA'
              mask = sly.AlphaMask.base64_2_data(uncompressed_string)
              print(mask)
        """
        try:
            z = zlib.decompress(base64.b64decode(s))
        except zlib.error:
            # If the string is not compressed, we'll not use zlib.
            img = Image.open(io.BytesIO(base64.b64decode(s)))
            return np.array(img)
        n = np.frombuffer(z, np.uint8)

        imdecoded = cv2.imdecode(n, cv2.IMREAD_GRAYSCALE)  # pylint: disable=no-member
        if (len(imdecoded.shape) == 3) and (imdecoded.shape[2] == 4):
            mask = imdecoded[:, :, 3]  # pylint: disable=unsubscriptable-object
        if (len(imdecoded.shape) == 3) and (imdecoded.shape[2] == 1):
            mask = imdecoded[:, :, 0] # pylint: disable=unsubscriptable-object
        elif len(imdecoded.shape) == 2:
            mask = imdecoded
        else:
            raise RuntimeError("Wrong internal mask format.")
        return mask

    @classmethod
    def allowed_transforms(cls):
        """allowed_transforms"""
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.bitmap import Bitmap
        from supervisely.geometry.polygon import Polygon
        from supervisely.geometry.rectangle import Rectangle

        return [AnyGeometry, Bitmap, Polygon, Rectangle]

    @classmethod
    def from_path(cls, path: str) -> AlphaMask:
        """
        Read alpha_channel from image by path.

        :param path: Path to image
        :type path: str
        :return: AlphaMask
        :rtype: AlphaMask
        """
        img = cv2.imread(path, cv2.IMREAD_UNCHANGED)  # pylint: disable=no-member
        if len(img.shape) == 2:
            return AlphaMask(img)
        elif img.shape[2] == 1:
            return AlphaMask(img[:, :, 0])
        elif img.shape[2] == 4:
            img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)  # pylint: disable=no-member
            return AlphaMask(img[:, :, 3])
        else:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)  # pylint: disable=no-member
            return AlphaMask(img)



from supervisely.api.api import Api


class Chart:
    def __init__(self, task_id, api: Api, v_model, title, series_names=None,
                 smoothing=None, yrange=None, ydecimals=None, xdecimals=None):
        self._task_id = task_id
        self._api = api
        self._v_model = v_model
        if self._v_model.startswith("data.") is False:
            new_v_model = v_model.replace('state.', 'data.', 1)
            raise KeyError(f"Data for this widget has to be stored in data field. "
                           f"Please, change value of v_model argument "
                           f"from {v_model} to {new_v_model} manually and also check html template")

        self._title = title
        if series_names is None:
            self._series_names = [title]
        else:
            self._series_names = series_names
        self._smoothing = smoothing
        self._yrange = yrange
        self._ydecimals = ydecimals
        self._xdecimals = xdecimals

        self._series = []
        for name in self._series_names:
            self._series.append({
                "name": name,
                "data": []
            })

    def to_json(self):
        result = {
            "options": {
                "title": self._title,
                # "groupKey": "my-synced-charts",
            },
            "series": self._series
        }
        if self._smoothing is not None:
            result["options"]["smoothingWeight"] = self._smoothing
        if self._yrange is not None:
            result["options"]["yaxisInterval"] = self._yrange
        if self._ydecimals is not None:
            result["options"]["decimalsInFloat"] = self._ydecimals
        if self._xdecimals is not None:
            result["options"]["xaxisDecimalsInFloat"] = self._xdecimals
        return result

    def init_data(self, data: dict):
        data[self._v_model.split(".")[1]] = self.to_json()

    def get_field(self, x, y, series_name=None):
        if series_name is None:
            series_name = self._title
        try:
            index = self._series_names.index(series_name)
        except ValueError:
            raise ValueError(f"Series '{series_name}' not found in {self._series_names}")
        return {"field": f"{self._v_model}.series[{index}].data", "payload": [[x, y]], "append": True}

    def append(self, x, y, series_name=None):
        field = self.get_field(x, y, series_name)
        self._api.app.set_fields(self._task_id, [field])


from supervisely.app.widgets import Widget
from supervisely.annotation.tag_meta import TagMeta


class TagMetaView(Widget):
    def __init__(
        self,
        tag_meta: TagMeta,
        show_type_text: bool = True,
        limit_long_names: bool = False,
        widget_id: str = None,
    ):
        self._tag_meta = tag_meta
        self._show_type_text = show_type_text
        self._limit_long_names = limit_long_names
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        res = self._tag_meta.to_json()
        res["limit_long_names"] = self._limit_long_names
        res["type_text"] = None
        if self._show_type_text is True:
            res["type_text"] = self._tag_meta.value_type.upper()
        return res

    def get_json_state(self):
        return None



# coding: utf-8
"""get list of plugins for supervisely team"""

from __future__ import annotations

from typing import Dict, List, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleApi


class PluginApi(ModuleApi):
    """
    API for working with plugins. :class:`PluginApi<PluginApi>` object is immutable.

    :param api: API connection to the server
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        team_id = 8
        plugin_info = api.plugin.get_list(team_id)
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple PluginInfo information about Plugin.

        :Example:

         .. code-block:: python

            PluginInfo(id=3,
                       name='DTL',
                       description='Allows to combine datasets, to make class mapping, filter objects and images, apply auto augmentations and so on ...',
                       type='dtl',
                       default_version='latest',
                       docker_image='docker.deepsystems.io/supervisely/five/dtl',
                       readme='# Data Transformation Language (DTL)...',
                       configs=[],
                       versions=['lately', 'docs', ...],
                       created_at='2020-03-30T09:17:36.000Z',
                       updated_at='2020-04-23T06:26:29.000Z')
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.TYPE,
            ApiField.DEFAULT_VERSION,
            ApiField.DOCKER_IMAGE,
            ApiField.README,
            ApiField.CONFIGS,
            ApiField.VERSIONS,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **PluginInfo**.
        """
        return "PluginInfo"

    def get_list(
        self, team_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        Get list of plugins in the Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param filters: List of params to sort output Plugins.
        :type filters: List[dict], optional
        :return: List of Plugins with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_id = 8
            plugin_info = api.plugin.get_list(team_id)

            plugin_list_filter = api.plugin.get_list(team_id, filters=[{'field': 'name', 'operator': '=', 'value': 'Images'}])
        """
        return self.get_list_all_pages(
            "plugins.list", {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters or []}
        )

    def get_info_by_id(self, team_id: int, plugin_id: int) -> NamedTuple:
        """
        Get Plugin information by ID.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param plugin_id: Plugin ID in Supervisely.
        :type plugin_id: int
        :return: Information about Plugin. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            plugin_info = api.plugin.get_info_by_id(8, 3)
            print(plugin_info)
            # Output: PluginInfo(id=3,
            #                    name='DTL',
            #                    description='Allows to combine datasets, to make class mapping, filter objects and images, apply auto augmentations and so on ...',
            #                    type='dtl',
            #                    default_version='latest',
            #                    docker_image='docker.deepsystems.io/supervisely/five/dtl',
            #                    readme='# Data Transformation Language (DTL)...',
            #                    configs=[],
            #                    versions=['lately', 'docs', ...],
            #                    created_at='2020-03-30T09:17:36.000Z',
            #                    updated_at='2020-04-23T06:26:29.000Z')
        """
        filters = [{"field": ApiField.ID, "operator": "=", "value": plugin_id}]
        return self._get_info_by_filters(team_id, filters)



import io
import os
import tarfile
import tempfile
import time
from datetime import datetime
from typing import List, NamedTuple, Tuple, Union

import requests
import zstd

from supervisely._utils import logger
from supervisely.api.module_api import ApiField, ModuleApiBase
from supervisely.api.project_api import ProjectInfo
from supervisely.io import json
from supervisely.io.fs import remove_dir, silent_remove


class VersionInfo(NamedTuple):
    """ """

    id: int
    project_id: int
    created_by: int
    team_file_id: int
    version: int
    description: str
    status: str
    created_at: str
    updated_at: str
    project_updated_at: str
    team_id: int


class DataVersion(ModuleApiBase):

    def __init__(self, api):
        """
        Class for managing project versions.
        """
        from supervisely import Api

        self._api: Api = api
        self.__storage_dir: str = "/system/versions/"
        self.__version_format: str = "v1.0.0"
        self.project_info = None
        self.project_dir = None
        self.versions_path = None
        self.versions = None

    @staticmethod
    def info_sequence():
        """
        NamedTuple VersionInfo with API Fields containing information about Project Version.

        """

        return [
            ApiField.ID,
            ApiField.PROJECT_ID,
            ApiField.CREATED_BY_ID,
            ApiField.TEAM_FILE_ID,
            ApiField.VERSION,
            ApiField.DESCRIPTION,
            ApiField.STATUS,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.PROJECT_UPDATED_AT,
            ApiField.TEAM_ID,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **VersionInfo**.
        """
        return "VersionInfo"

    def initialize(self, project_info: Union[ProjectInfo, int]):
        """
        Initialize project versions.

        :param project_info: ProjectInfo object or project ID
        :type project_info: Union[ProjectInfo, int]
        """
        if isinstance(project_info, int):
            project_info = self._api.project.get_info_by_id(project_info)
        self.project_info: ProjectInfo = project_info
        self.project_dir: str = os.path.join(self.__storage_dir, str(self.project_info.id))
        self.versions_path: str = os.path.join(self.project_dir, "versions.json")
        self.versions: dict = self.get_map(self.project_info, do_initialization=False)
        if self.project_info.version is None:
            self._create_warning_system_file()

    def get_list(self, project_id: int, filters: List = None) -> List[VersionInfo]:
        """
        Get list of project versions.

        :param project_id: Project ID
        :type project_id: int
        :param filters: Filters
        :type filters: Optional[List]
        :return: List of project versions
        :rtype: List[VersionInfo]
        """
        data = {ApiField.PROJECT_ID: project_id}
        if filters:
            data[ApiField.FILTER] = filters

        return self.get_list_all_pages("projects.versions.list", data)

    def get_id_by_number(self, project_id: int, version_num: int) -> int:
        """
        Get version ID by version number.

        :param project_id: Project ID
        :type project_id: int
        :param version_num: Version number
        :type version_num: int
        :return: Version ID
        :rtype: int or None
        """
        filter = [
            {
                ApiField.FIELD: ApiField.VERSION,
                ApiField.OPERATOR: "=",
                ApiField.VALUE: int(version_num),
            }
        ]
        versions = self.get_list(project_id, filters=filter)
        if len(versions) > 0:
            return versions[0].id
        return None

    def get_map(self, project_info: Union[ProjectInfo, int], do_initialization: bool = True):
        """
        Get project versions map from storage.

        :param project_info: ProjectInfo object or project ID
        :type project_info: Union[ProjectInfo, int]
        :param do_initialization: Initialize project versions. Set to False for internal use.
        :type do_initialization: bool
        :return: Project versions
        :rtype: dict
        """
        if do_initialization:
            self.initialize(project_info)

        try:
            versions = self._api.file.get_json_file_content(
                self.project_info.team_id, self.versions_path
            )
            versions = versions if versions else {}
        except FileNotFoundError:
            versions = {"format": self.__version_format}
        return versions

    def set_map(self, project_info: Union[ProjectInfo, int], initialize: bool = True):
        """
        Save project versions map to storage.

        :param project_info: ProjectInfo object or project ID
        :type project_info: Union[ProjectInfo, int]
        :param initialize: Initialize project versions. Set to False for internal use.
        :type initialize: bool
        :return: None
        """

        if initialize:
            self.initialize(project_info)
        temp_dir = tempfile.mkdtemp()
        local_versions = os.path.join(temp_dir, "versions.json")
        json.dump_json_file(self.versions, local_versions)
        file_info = self._api.file.upload(
            self.project_info.team_id, local_versions, self.versions_path
        )
        if file_info is None:
            raise RuntimeError("Failed to save versions")

        remove_dir(temp_dir)

    def create(
        self,
        project_info: Union[ProjectInfo, int],
        version_title: str = None,
        version_description: str = None,
    ) -> int:
        """
        Create a new project version.
        Returns the ID of the new version.
        If the project is already on the latest version, returns the latest version ID.
        If the project version cannot be created, returns None.

        :param project_info: ProjectInfo object or project ID
        :type project_info: Union[ProjectInfo, int]
        :param version_title: Version title
        :type version_title: str
        :param version_description: Version description
        :type version_description: str
        :return: Version ID
        :rtype: int
        """
        if isinstance(project_info, int):
            project_info = self._api.project.get_info_by_id(project_info)

        if (
            "app.supervise.ly" in self._api.server_address
            or "app.supervisely.com" in self._api.server_address
        ):
            if self._api.team.get_info_by_id(project_info.team_id).usage.plan == "free":
                logger.warning(
                    "Project versioning is not available for teams with Free plan. Please upgrade to Pro to enable versioning."
                )
                return None

        self.initialize(project_info)
        path = self._generate_save_path()
        latest = self._get_latest_id()
        try:
            version_id, commit_token = self.reserve(project_info.id)
        except Exception as e:
            logger.error(f"Failed to reserve version. Exception: {e}")
            return None
        if version_id is None and commit_token is None:
            return latest
        try:
            file_info = self._compress_and_upload(path)
            self.versions[version_id] = {
                "path": path,
                "updated_at": project_info.updated_at,
                "previous": latest,
                "number": int(self.versions[str(latest)]["number"]) + 1 if latest else 1,
            }
            self.versions["latest"] = version_id
            self.set_map(project_info, initialize=False)
            self.commit(
                version_id,
                commit_token,
                project_info.updated_at,
                file_info.id,
                title=version_title,
                description=version_description,
            )
            return version_id
        except Exception as e:
            if self.cancel_reservation(version_id, commit_token):
                logger.error(f"Version creation failed. Reservation was cancelled. Exception: {e}")
            else:
                logger.error(
                    f"Failed to cancel reservation when handling exception. You can cancel your reservation on the web under the Versions tab of the project. Exception: {e}"
                )
            return None

    def commit(
        self,
        version_id: int,
        commit_token: str,
        updated_at: str,
        file_id: int,
        title: str = None,
        description: str = None,
    ):
        """
        Commit project version.
        This method is used to finalize the version creation process.
        Requires active reservation.
        You must call this method after creating project version backup and setting version map

        :param version_id: Version ID
        :type version_id: int
        :param commit_token: Commit token
        :type commit_token: str
        :param updated_at: Updated at timestamp
        :type updated_at: str
        :param file_id: File ID
        :type file_id: int
        :param title: Version title
        :type title: str
        :param description: Version description
        :type description: str
        :return: None
        """
        body = {
            ApiField.ID: version_id,
            ApiField.COMMIT_TOKEN: commit_token,
            ApiField.PROJECT_UPDATED_AT: updated_at,
            ApiField.TEAM_FILE_ID: file_id,
        }
        if title:
            body[ApiField.TITLE] = title
        if description:
            body[ApiField.DESCRIPTION] = description

        response = self._api.post("projects.versions.commit", body)
        commit_info = response.json()
        if not commit_info.get("success"):
            raise RuntimeError("Failed to commit version")

    def reserve(self, project_id: int, retries: int = 6) -> Tuple[int, str]:
        """
        Reserve project version.
        This method is used before backing up a version to prevent another attempt to create a version at the same time.
        The first delay of retry is 2 seconds, which doubles with each subsequent attempt.

        :param project_id: Project ID
        :type project_id: int
        :param retries: Number of attempts to reserve version
        :type retries: int
        :return: Version ID and commit token
        :rtype: Tuple[int, str]
        """
        retry_delay = 2  # seconds
        max_delay = retry_delay * 2**retries

        while True:
            try:
                response = self._api.post(
                    "projects.versions.reserve", {ApiField.PROJECT_ID: project_id}
                )
                reserve_info = response.json()
                return reserve_info.get(ApiField.ID), reserve_info.get(ApiField.COMMIT_TOKEN)

            except requests.exceptions.HTTPError as e:
                if e.response.json().get("details", {}).get("useExistingVersion"):
                    version_id = e.response.json().get("details", {}).get("version").get("id")
                    version = e.response.json().get("details", {}).get("version").get("version")
                    logger.info(
                        f"No changes to the project since the last version '{version}' with ID '{version_id}'"
                    )
                    return (None, None)
                elif "is already committing" in e.response.json().get("details", {}).get("message"):
                    if retry_delay >= max_delay:
                        raise RuntimeError(
                            "Failed to reserve version. Another process is already committing a version. Maximum number of attempts reached."
                        )
                    version = e.response.json().get("details", {}).get("version").get("version")
                    time.sleep(retry_delay)
                    retry_delay *= 2

    def cancel_reservation(self, version_id: int, commit_token: str):
        """
        Cancel version reservation for a project.

        :param version_id: Version ID
        :type version_id: int
        :param commit_token: Commit token
        :type commit_token: str
        :return: True if reservation was cancelled, False otherwise
        """
        response = self._api.post(
            "projects.versions.cancel-reservation",
            {ApiField.ID: version_id, ApiField.COMMIT_TOKEN: commit_token},
        )
        reserve_info = response.json()
        return True if reserve_info.get("success") else False

    def restore(
        self,
        project_info: Union[ProjectInfo, int],
        version_id: int = None,
        version_num: int = None,
        skip_missed_entities: bool = False,
    ) -> ProjectInfo:
        """
        Restore project to a specific version.
        Version can be specified by ID or number.

        :param project_info: ProjectInfo object or project ID
        :type project_info: Union[ProjectInfo, int]
        :param version_id: Version ID
        :type version_id: Optional[int]
        :param version_num: Version number
        :type version_num: Optional[int]
        :param skip_missed_entities: Skip missed Images
        :type skip_missed_entities: bool, default False
        :return: ProjectInfo object of the restored project
        :rtype: ProjectInfo or None
        """
        from supervisely.project.project import Project

        if version_id is None and version_num is None:
            raise ValueError("Either version_id or version_num must be provided")

        self.initialize(project_info)

        if version_num:
            version_id = None
            for key, value in self.versions.items():
                # pylint: disable=no-member
                if isinstance(value, dict) and value.get("number") == version_num:
                    version_id = key
                    break
            if version_id is None:
                raise ValueError(f"Version {version_num} does not exist")
        else:
            if str(version_id) not in self.versions:
                raise ValueError(f"Version {version_id} does not exist")
            version_num = self.versions[str(version_id)]["number"]
        updated_at = self.versions[str(version_id)]["updated_at"]
        backup_files = self.versions[str(version_id)]["path"]

        if updated_at == self.project_info.updated_at:
            logger.warning(
                f"Project is already on version {version_num} with the same updated_at timestamp"
            )
            return

        if backup_files is None:
            logger.warning(
                f"Project can't be restored to version {version_num} because it doesn't have restore point."
            )
            return

        bin_io = self._download_and_extract(backup_files)
        new_project_info = Project.upload_bin(
            self._api,
            bin_io,
            self.project_info.workspace_id,
            skip_missed=skip_missed_entities,
        )
        return new_project_info

    def _create_warning_system_file(self):
        """
        Create a file in the system directory to indicate that you cannot manually modify its contents.

        Path = /system/DO_NOT_DELETE_ANYTHING_HERE.txt

        """
        warning_file = "/system/DO_NOT_DELETE_ANYTHING_HERE.txt"
        if not self._api.file.exists(self.project_info.team_id, warning_file, recursive=False):
            temp_file = tempfile.NamedTemporaryFile(delete=False)
            with open(temp_file.name, "w") as f:
                f.write("This directory is managed by Supervisely. Do not modify its contents.")
            self._api.file.upload(self.project_info.team_id, temp_file.name, warning_file)

    def _download_and_extract(self, path: str) -> io.BytesIO:
        """
        Download and extract version data to memory.

        :param path: Path to the version file
        :type path: str
        :return: Binary IO object with extracted file
        :rtype: io.BytesIO
        """
        temp_dir = tempfile.mkdtemp()
        local_path = os.path.join(temp_dir, "download.tar.zst")
        try:
            self._api.file.download(self.project_info.team_id, path, local_path)
            with open(local_path, "rb") as zst:
                decompressed_data = zstd.decompress(zst.read())
            with tarfile.open(fileobj=io.BytesIO(decompressed_data)) as tar:
                file = tar.extractfile("version.bin")
                if not file:
                    raise RuntimeError("version.bin not found in the archive")
                data = file.read()
                bin_io = io.BytesIO(data)
                return bin_io
        except Exception as e:
            raise RuntimeError(f"Failed to extract version: {e}")
        finally:
            remove_dir(temp_dir)

    def _generate_save_path(self):
        """
        Generate a path for the new version archive where it will be saved in the Team Files.
        Archive format: {timestamp}.tar.zst

        :return: Path for the new version archive
        :rtype: str
        """
        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        path = os.path.join(self.project_dir, timestamp + ".tar.zst")
        return path

    def _get_latest_id(self):
        """
        Get the ID of the latest version from the versions map (versions.json).
        """
        latest = self.versions.get("latest", None)
        if not latest:
            return None
        return latest

    def _compress_and_upload(self, path: str) -> dict:
        """
        Save project in binary format in archive to the Team Files.
        Binary file name: version.bin

        :param changes: Changes between current and previous version
        :type changes: bool
        :return: File info
        :rtype: dict
        """
        from supervisely.project.project import Project

        temp_dir = tempfile.mkdtemp()

        data = Project.download_bin(
            self._api, self.project_info.id, batch_size=200, return_bytesio=True
        )
        data.seek(0)
        info = tarfile.TarInfo(name="version.bin")
        info.size = len(data.getvalue())
        chunk_size = 1024 * 1024 * 50  # 50 MiB
        tar_data = io.BytesIO()

        # Create a tarfile object that writes into the BytesIO object
        with tarfile.open(fileobj=tar_data, mode="w") as tar:
            tar.addfile(tarinfo=info, fileobj=data)
        data.close()
        # Reset the BytesIO object's cursor to the beginning
        tar_data.seek(0)
        zst_archive_path = os.path.join(temp_dir, "download.tar.zst")

        with open(zst_archive_path, "wb") as zst:
            while True:
                chunk = tar_data.read(chunk_size)
                if not chunk:
                    break
                zst.write(zstd.compress(chunk))
        file_info = self._api.file.upload(self.project_info.team_id, zst_archive_path, path)
        tar_data.close()
        remove_dir(temp_dir)
        return file_info



import supervisely.nn.artifacts as artifacts
import supervisely.nn.inference as inference
from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts, TrainInfo
from supervisely.nn.prediction_dto import (
    Prediction,
    PredictionBBox,
    PredictionCuboid3d,
    PredictionKeypoints,
    PredictionMask,
    PredictionSegmentation,
)
